"""Tests for text embedding module"""

# Standard
from typing import List, Tuple
import os
import tempfile

# Third Party
from pytest import approx
from torch.backends import mps
import numpy as np
import pytest
import torch

# First Party
from caikit.core import ModuleConfig
from caikit.interfaces.common.data_model.vectors import ListOfVector1D
from caikit.interfaces.nlp.data_model import (
    EmbeddingResult,
    RerankResult,
    RerankResults,
    RerankScore,
    RerankScores,
)

# Local
from caikit_nlp.modules.text_embedding import EmbeddingModule, utils
from caikit_nlp.modules.text_embedding.embedding import (
    get_sample_start_indexes,
    sum_token_count,
)
from tests.fixtures import SEQ_CLASS_MODEL

## Setup ########################################################################

# Bootstrapped sequence classification model for reuse across tests
# .bootstrap is tested separately in the first test
BOOTSTRAPPED_MODEL = EmbeddingModule.bootstrap(SEQ_CLASS_MODEL)

# Token counts:
# All expected token counts were calculated with reference to the
# `BertForSequenceClassification` model. Each model's tokenizer behaves differently
# which can lead to the expected token counts being invalid.

INPUT = "The quick brown fox jumps over the lazy dog."
INPUT_TOKEN_COUNT = 36 + 2  # [CLS] Thequickbrownfoxjumpsoverthelazydog. [SEP]

MANY_INPUTS = [
    "The quick brown fox jumps over the lazy dog.",
    "But I must explain to you how all this mistaken idea.",
    "No one rejects or dislikes.",
]

QUERY = "What is foo bar?"
QUERY_TOKEN_COUNT = 13 + 2  # [CLS] Whatisfoobar? [SEP]

QUERIES: List[str] = [
    "Who is foo?",
    "Where is the bar?",
]
QUERIES_TOKEN_COUNT = (9 + 2) + (
    14 + 2
)  # [CLS] Whoisfoo? [SEP], [CLS] Whereisthebar? [SEP]


# These are used to test that documents can handle different types in and out
TYPE_KEYS = "str_test", "int_test", "float_test", "nested_dict_test"

DOCS = [
    {
        "text": "foo",
        "title": "title or whatever",
        "str_test": "test string",
        "int_test": 1,
        "float_test": 1.234,
        "score": 99999,
        "nested_dict_test": {"deep1": 1, "deep string": "just testing"},
    },
    {
        "_text": "bar",
        "title": "title 2",
    },
    {
        "text": "foo and bar",
    },
    {
        "_text": "Where is the bar",
        "another": "something else",
    },
]

# The `text` and `_text` keys are extracted from DOCS as input to the tokenizer
# [CLS] foo [SEP], [CLS] bar [SEP], [CLS] fooandbar [SEP], [CLS] Whereisthebar [SEP]
DOCS_TOKEN_COUNT = (3 + 2) + (3 + 2) + (9 + 2) + (13 + 2)

# Use text or _text from DOCS for our test sentences
SENTENCES = [d.get("text", d.get("_text")) for d in DOCS]

# [CLS] foo [SEP], [CLS] bar [SEP], [CLS] fooandbar [SEP], [CLS] Whereisthebar [SEP]
SENTENCES_TOKEN_COUNT = (3 + 2) + (3 + 2) + (9 + 2) + (13 + 2)

## Tests ########################################################################


@pytest.fixture(scope="module", name="loaded_model")
def fixture_loaded_model(tmp_path_factory):
    models_dir = tmp_path_factory.mktemp("models")
    model_path = str(models_dir / "model_id")
    BOOTSTRAPPED_MODEL.save(model_path)
    model = EmbeddingModule.load(model_path)
    return model


def _assert_is_expected_vector(vector):
    assert isinstance(vector.data.values[0], np.float32)
    assert len(vector.data.values) == 32
    # Just testing a few values for readability
    assert approx(vector.data.values[0]) == 0.3244932293891907
    assert approx(vector.data.values[1]) == -0.4934631288051605
    assert approx(vector.data.values[2]) == 0.5721234083175659


def _assert_is_expected_embedding_result(actual):
    assert isinstance(actual, EmbeddingResult)
    vector = actual.result
    _assert_is_expected_vector(vector)


def _assert_is_expected_embeddings_results(actual):
    assert isinstance(actual, ListOfVector1D)
    _assert_is_expected_vector(actual.vectors[0])


def test_bootstrap():
    assert isinstance(
        EmbeddingModule.bootstrap(SEQ_CLASS_MODEL), EmbeddingModule
    ), "bootstrap error"


def _assert_types_found(types_found):
    assert type(types_found["str_test"]) == str, "passthru str value type check"
    assert type(types_found["int_test"]) == int, "passthru int value type check"
    assert type(types_found["float_test"]) == float, "passthru float value type check"
    assert (
        type(types_found["nested_dict_test"]) == dict
    ), "passthru nested dict value type check"


def _assert_valid_scores(scores, type_tests={}):
    for score in scores:
        assert isinstance(score, RerankScore)
        assert isinstance(score.score, float)
        assert isinstance(score.index, int)
        assert isinstance(score.text, str)

        document = score.document
        assert isinstance(document, dict)
        assert document == DOCS[score.index]

        # Test document key named score (None or 9999) is independent of the result score
        assert score.score != document.get(
            "score"
        ), "unexpected passthru score same as result score"

        # Gather various type test values when we have them
        for k, v in document.items():
            if k in TYPE_KEYS:
                type_tests[k] = v

    return type_tests


def test_bootstrap_model(loaded_model):
    assert isinstance(BOOTSTRAPPED_MODEL, EmbeddingModule), "bootstrap model type"
    assert (
        BOOTSTRAPPED_MODEL.model.__class__.__name__ == "SentenceTransformer"
    ), "bootstrap model class name"
    # worth noting that bootstrap does not wrap, but load does
    assert (
        loaded_model.model.__class__.__name__ == "SentenceTransformerWithTruncate"
    ), "loaded model class name"


def test_save_load_and_run():
    """Check if we can load and run a saved model successfully"""
    model_id = "model_id"
    with tempfile.TemporaryDirectory(suffix="-1st") as model_dir:
        model_path = os.path.join(model_dir, model_id)
        BOOTSTRAPPED_MODEL.save(model_path)
        new_model = EmbeddingModule.load(model_path)

    assert isinstance(new_model, EmbeddingModule), "save and load error"
    assert new_model != BOOTSTRAPPED_MODEL, "did not load a new model"

    # Use run_embedding just to make sure this new model is usable
    result = new_model.run_embedding(text=INPUT)
    _assert_is_expected_embedding_result(result)


def test_public_model_info():
    """Check if we can get model info successfully"""
    model_id = "model_id"
    with tempfile.TemporaryDirectory(suffix="-1st") as model_dir:
        model_path = os.path.join(model_dir, model_id)
        BOOTSTRAPPED_MODEL.save(model_path)
        new_model = EmbeddingModule.load(model_path)

    result = new_model.public_model_info
    assert "max_seq_length" in result
    assert "sentence_embedding_dimension" in result
    assert type(result["max_seq_length"]) is int
    assert type(result["sentence_embedding_dimension"]) is int


@pytest.mark.parametrize(
    "model_path", ["", " ", " " * 100], ids=["empty", "space", "spaces"]
)
def test_save_value_checks(model_path):
    with pytest.raises(ValueError):
        BOOTSTRAPPED_MODEL.save(model_path)


@pytest.mark.parametrize(
    "model_path",
    ["..", "../" * 100, "/", ".", " / ", " . "],
)
def test_save_exists_checks(model_path):
    """Tests for model paths are always existing dirs that should not be clobbered"""
    with pytest.raises(FileExistsError):
        BOOTSTRAPPED_MODEL.save(model_path)


def test_second_save_hits_exists_check():
    """Using a new path the first save should succeed but second fails"""
    model_id = "model_id"
    with tempfile.TemporaryDirectory(suffix="-2nd") as model_dir:
        model_path = os.path.join(model_dir, model_id)
        BOOTSTRAPPED_MODEL.save(model_path)
        with pytest.raises(FileExistsError):
            BOOTSTRAPPED_MODEL.save(model_path)


@pytest.mark.parametrize("model_path", [None, {}, object(), 1], ids=type)
def test_save_type_checks(model_path):
    with pytest.raises(TypeError):
        BOOTSTRAPPED_MODEL.save(model_path)


def test_load_without_artifacts():
    """Test coverage for the error message when config has no artifacts to load"""
    with pytest.raises(ValueError):
        EmbeddingModule.load(ModuleConfig({}))


def test_run_embedding_type_check(loaded_model):
    """Input cannot be a list"""
    with pytest.raises(TypeError):
        loaded_model.run_embedding([INPUT])
        pytest.fail("Should not reach here")


def test_run_embedding(loaded_model):
    res = loaded_model.run_embedding(text=INPUT)
    _assert_is_expected_embedding_result(res)
    assert res.input_token_count == INPUT_TOKEN_COUNT


def test_run_embeddings_str_type(loaded_model):
    """Supposed to be a list, gets fixed automatically."""
    res = loaded_model.run_embeddings(texts=INPUT)
    assert isinstance(res.results.vectors, list)
    assert len(res.results.vectors) == 1


def test_run_embeddings(loaded_model):
    res = loaded_model.run_embeddings(texts=[INPUT])
    assert isinstance(res.results.vectors, list)
    _assert_is_expected_embeddings_results(res.results)
    assert res.input_token_count == INPUT_TOKEN_COUNT


@pytest.mark.parametrize(
    "query,docs,top_n",
    [
        (["test list"], DOCS, None),
        (None, DOCS, 1234),
        (False, DOCS, 1234),
        (QUERY, {"testdict": "not list"}, 1234),
        (QUERY, DOCS, "topN string is not an integer or None"),
    ],
)
def test_run_rerank_query_type_error(query, docs, top_n, loaded_model):
    """test for type checks matching task/run signature"""
    with pytest.raises(TypeError):
        loaded_model.run_rerank_query(query=query, documents=docs, top_n=top_n)
        pytest.fail("Should not reach here.")


def test_run_rerank_query_no_type_error(loaded_model):
    """no type error with list of string queries and list of dict documents"""
    res = loaded_model.run_rerank_query(query=QUERY, documents=DOCS, top_n=1)
    assert res.input_token_count == QUERY_TOKEN_COUNT + DOCS_TOKEN_COUNT


@pytest.mark.parametrize(
    "top_n, expected",
    [
        (1, 1),
        (2, 2),
        (None, len(DOCS)),
        (-1, len(DOCS)),
        (0, len(DOCS)),
        (9999, len(DOCS)),
    ],
)
def test_run_rerank_query_top_n(top_n, expected, loaded_model):
    res = loaded_model.run_rerank_query(query=QUERY, documents=DOCS, top_n=top_n)
    assert isinstance(res, RerankResult)
    assert len(res.result.scores) == expected
    assert res.input_token_count == QUERY_TOKEN_COUNT + DOCS_TOKEN_COUNT


def test_run_rerank_query_no_query(loaded_model):
    with pytest.raises(TypeError):
        loaded_model.run_rerank_query(query=None, documents=DOCS, top_n=99)


def test_run_rerank_query_zero_docs(loaded_model):
    """No empty doc list therefore result is zero result scores"""
    with pytest.raises(ValueError):
        loaded_model.run_rerank_query(query=QUERY, documents=[], top_n=99)


def test_run_rerank_query(loaded_model):
    res = loaded_model.run_rerank_query(query=QUERY, documents=DOCS)
    assert isinstance(res, RerankResult)

    scores = res.result.scores
    assert isinstance(scores, list)
    assert len(scores) == len(DOCS)

    types_found = _assert_valid_scores(scores)
    _assert_types_found(types_found)
    assert res.input_token_count == QUERY_TOKEN_COUNT + DOCS_TOKEN_COUNT


@pytest.mark.parametrize(
    "queries,docs", [("test string", DOCS), (QUERIES, {"testdict": "not list"})]
)
def test_run_rerank_queries_type_error(queries, docs, loaded_model):
    """type error check ensures params are lists and not just 1 string or just one doc (for example)"""
    with pytest.raises(TypeError):
        loaded_model.run_rerank_queries(queries=queries, documents=docs)
        pytest.fail("Should not reach here.")


def test_run_rerank_queries_no_type_error(loaded_model):
    """no type error with list of string queries and list of dict documents"""
    res = loaded_model.run_rerank_queries(queries=QUERIES, documents=DOCS, top_n=99)
    assert res.input_token_count == QUERIES_TOKEN_COUNT + DOCS_TOKEN_COUNT


@pytest.mark.parametrize(
    "top_n, expected",
    [
        (1, 1),
        (2, 2),
        (None, len(DOCS)),
        (-1, len(DOCS)),
        (0, len(DOCS)),
        (9999, len(DOCS)),
    ],
)
def test_run_rerank_queries_top_n(top_n, expected, loaded_model):
    """no type error with list of string queries and list of dict documents"""
    res = loaded_model.run_rerank_queries(queries=QUERIES, documents=DOCS, top_n=top_n)
    assert isinstance(res, RerankResults)
    assert len(res.results) == len(QUERIES)
    for result in res.results:
        assert len(result.scores) == expected
    assert res.input_token_count == QUERIES_TOKEN_COUNT + DOCS_TOKEN_COUNT


@pytest.mark.parametrize(
    "queries, docs",
    [
        ([], DOCS),
        (QUERIES, []),
        ([], []),
    ],
    ids=["no queries", "no docs", "no queries and no docs"],
)
def test_run_rerank_queries_no_queries_or_no_docs(queries, docs, loaded_model):
    """No queries and/or no docs therefore result is zero results"""

    with pytest.raises(ValueError):
        loaded_model.run_rerank_queries(queries=queries, documents=docs, top_n=9)


def test_run_rerank_queries(loaded_model):
    top_n = 2
    rerank_result = loaded_model.run_rerank_queries(
        queries=QUERIES, documents=DOCS, top_n=top_n
    )
    assert isinstance(rerank_result, RerankResults)

    results = rerank_result.results
    assert isinstance(results, list)
    assert len(results) == 2 == len(QUERIES)  # 2 queries yields 2 result(s)

    types_found = {}  # Gather the type tests from any of the results

    for result in results:
        assert isinstance(result, RerankScores)
        scores = result.scores
        assert isinstance(scores, list)
        assert len(scores) == top_n
        types_found = _assert_valid_scores(scores, types_found)

    # Make sure our document fields of different types made it in/out ok
    _assert_types_found(types_found)
    assert rerank_result.input_token_count == QUERIES_TOKEN_COUNT + DOCS_TOKEN_COUNT


def test_run_sentence_similarity(loaded_model):
    res = loaded_model.run_sentence_similarity(
        source_sentence=QUERY, sentences=SENTENCES
    )
    scores = res.result.scores
    assert len(scores) == len(SENTENCES)
    for score in scores:
        assert isinstance(score, float)
    assert res.input_token_count == QUERY_TOKEN_COUNT + SENTENCES_TOKEN_COUNT


def test_run_sentence_similarities(loaded_model):
    res = loaded_model.run_sentence_similarities(
        source_sentences=QUERIES, sentences=SENTENCES
    )
    results = res.results
    assert len(results) == len(QUERIES)
    for result in results:
        scores = result.scores
        assert len(scores) == len(SENTENCES)
        for score in scores:
            assert isinstance(score, float)
    assert res.input_token_count == QUERIES_TOKEN_COUNT + SENTENCES_TOKEN_COUNT


@pytest.mark.parametrize(
    "use_ipex, device, expected",
    [
        (True, "", None),
        (False, "", None),
        (True, None, None),
        (False, None, None),
        (False, "xpu", None),
        (True, "xpu", "xpu"),
        (True, "mps", None),
        (
            False,
            "mps",
            "mps" if mps.is_built() and mps.is_available() else None,
        ),
    ],
)
def test__select_device(use_ipex, device, expected):
    assert EmbeddingModule._select_device(use_ipex, device) == expected


@pytest.mark.parametrize(
    "use_ipex, use_device, expected",
    [
        (True, None, "ipex"),
        (True, "mps", "ipex"),
        (False, "mps", mps),
        (False, None, "inductor"),
    ],
)
def test__get_backend(use_ipex, use_device, expected):
    # Make the Mac MPS test work depending on availability
    assert EmbeddingModule._get_backend(use_ipex, use_device) == expected


@pytest.mark.parametrize(
    "use_ipex",
    [None, "true", "True", "False", "false"],
)
def test__get_ipex(use_ipex):
    """Test that _get_ipex returns False instead of raising an exception.

    Assumes that when running tests, we won't have IPEX installed.
    """
    assert not EmbeddingModule._get_ipex(use_ipex)


def test__optimize():
    """Test that _optimize does nothing when disabled"""
    fake = "fake model"  # Will be returned as-is
    assert fake == EmbeddingModule._optimize(fake, False, "bogus", False, False)


@pytest.mark.parametrize("truncate_input_tokens", [0, 513])
def test__truncate_input_tokens_raises(truncate_input_tokens, loaded_model):
    model_max = loaded_model.model.max_seq_length

    too_long = "x " * (model_max - 1)  # This will go over
    over = model_max + 1
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.model.encode(
            sentences=[too_long], truncate_input_tokens=truncate_input_tokens
        )
    # Same behavior when implicit_truncation_errors is True (the default)
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.model.encode(
            sentences=[too_long],
            truncate_input_tokens=truncate_input_tokens,
            implicit_truncation_errors=True,
        )
    # Different behavior when implicit_truncation_errors is False -- no error raised!
    loaded_model.model.encode(
        sentences=[too_long],
        truncate_input_tokens=truncate_input_tokens,
        implicit_truncation_errors=False,
    )


def test__implicit_truncation(loaded_model):
    """Test that implicit truncation happens (when allowed)"""
    model_max = loaded_model.model.max_seq_length

    too_long = "x " * (model_max - 1)  # This will go over a little
    extra_long = (
        too_long
        + "more clever words that surely change the meaning of this text"
        * (model_max - 1)
    )

    # Allowed truncation using default tokens (0) and config to disable the error.
    res = loaded_model.model.encode(
        sentences=[too_long], truncate_input_tokens=0, implicit_truncation_errors=False
    )
    # Allowed truncation using model max
    res_extra_max = loaded_model.model.encode(
        sentences=[extra_long], truncate_input_tokens=loaded_model.model.max_seq_length
    )
    # Allowed truncation using -1 to just let the model do its thing
    res_extra_neg = loaded_model.model.encode(
        sentences=[extra_long], truncate_input_tokens=-1
    )

    # Demonstrating that when implicit truncation is allowed, sentence-transformers is quietly truncating at model max
    # The simple too_long string of x's, is equivalent to the string with significantly different extra text (truncated)
    assert np.allclose(res, res_extra_max)
    assert np.allclose(res, res_extra_neg)


def test_not_too_many_tokens(loaded_model):
    """Happy path for the endpoints using text that is not too many tokens."""

    model_max = loaded_model.model.max_seq_length

    ok = "x " * (model_max - 2)  # Subtract 2 for begin/end tokens

    # embedding(s)
    loaded_model.run_embedding(text=ok)
    loaded_model.run_embeddings(texts=[ok])

    # sentence similarity(ies) test both source_sentence and sentences
    loaded_model.run_sentence_similarity(source_sentence=ok, sentences=[ok])
    loaded_model.run_sentence_similarities(source_sentences=[ok], sentences=[ok])

    # reranker test both query and document text
    loaded_model.run_rerank_query(query=ok, documents=[{"text": ok}])
    loaded_model.run_rerank_queries(queries=[ok], documents=[{"text": ok}])


def test_too_many_tokens_default(loaded_model):
    """These endpoints raise an error when truncation would happen."""

    model_max = loaded_model.model.max_seq_length
    over = model_max + 1

    ok = "x " * (model_max - 2)  # Subtract 2 for begin/end tokens
    too_long = "x " * (model_max - 1)  # This will go over

    # embedding(s)
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_embedding(text=too_long)
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_embeddings(texts=[too_long])

    # sentence similarity(ies) test both source_sentence and sentences
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarity(source_sentence=too_long, sentences=[ok])
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarity(source_sentence=ok, sentences=[too_long])

    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarities(
            source_sentences=[too_long], sentences=[ok]
        )
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarities(
            source_sentences=[ok], sentences=[too_long]
        )

    # reranker test both query and document text
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_query(query=too_long, documents=[{"text": ok}])
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_query(query=ok, documents=[{"text": too_long}])

    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_queries(queries=[too_long], documents=[{"text": ok}])
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_queries(queries=[ok], documents=[{"text": too_long}])


@pytest.mark.parametrize("truncate_input_tokens", [0, 513])
def test_too_many_tokens_error_params(truncate_input_tokens, loaded_model):
    """truncate_input_tokens does not prevent these endpoints from raising an error.

    Test with 0 which uses the max model len (512) to determine truncation and raise error.
    Test with 513 (> 512) which detects truncation over 512 and raises an error.
    """

    model_max = loaded_model.model.max_seq_length
    over = model_max + 1

    ok = "x " * (model_max - 2)  # Subtract 2 for begin/end tokens
    too_long = "x " * (model_max - 1)  # This will go over

    # embedding(s)
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_embedding(
            text=too_long, truncate_input_tokens=truncate_input_tokens
        )
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_embeddings(
            texts=[too_long], truncate_input_tokens=truncate_input_tokens
        )

    # sentence similarity(ies) test both source_sentence and sentences
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarity(
            source_sentence=too_long,
            sentences=[ok],
            truncate_input_tokens=truncate_input_tokens,
        )
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarity(
            source_sentence=ok,
            sentences=[too_long],
            truncate_input_tokens=truncate_input_tokens,
        )

    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarities(
            source_sentences=[too_long],
            sentences=[ok],
            truncate_input_tokens=truncate_input_tokens,
        )
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_sentence_similarities(
            source_sentences=[ok],
            sentences=[too_long],
            truncate_input_tokens=truncate_input_tokens,
        )

    # reranker test both query and document text
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_query(
            query=too_long,
            documents=[{"text": ok}],
            truncate_input_tokens=truncate_input_tokens,
        )
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_query(
            query=ok,
            documents=[{"text": too_long}],
            truncate_input_tokens=truncate_input_tokens,
        )

    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_queries(
            queries=[too_long],
            documents=[{"text": ok}],
            truncate_input_tokens=truncate_input_tokens,
        )
    with pytest.raises(ValueError, match=f"({over} > {model_max})"):
        loaded_model.run_rerank_queries(
            queries=[ok],
            documents=[{"text": too_long}],
            truncate_input_tokens=truncate_input_tokens,
        )


@pytest.mark.parametrize("truncate_input_tokens", [-1, 99, 510, 511, 512])
def test_too_many_tokens_with_truncation_working(truncate_input_tokens, loaded_model):
    """truncate_input_tokens prevents these endpoints from raising an error when too many tokens.

    Test with -1 which lets the model do truncation instead of raising an error.
    Test with 99 (< 512 -2) which causes our code to do the truncation instead of raising an error.
    Test with 510 (512 -2) which causes our code to do the truncation instead of raising an error.
    511 and 512 also behave like 510. The value is allowed, but begin/end tokens will take space.
    """

    model_max = loaded_model.model.max_seq_length

    ok = "x " * (model_max - 2)  # Subtract 2 for begin/end tokens
    too_long = "x " * (model_max - 1)  # This will go over

    # embedding(s)
    loaded_model.run_embedding(
        text=too_long, truncate_input_tokens=truncate_input_tokens
    )
    loaded_model.run_embeddings(
        texts=[too_long], truncate_input_tokens=truncate_input_tokens
    )

    # sentence similarity(ies) test both source_sentence and sentences
    loaded_model.run_sentence_similarity(
        source_sentence=too_long,
        sentences=[ok],
        truncate_input_tokens=truncate_input_tokens,
    )
    loaded_model.run_sentence_similarity(
        source_sentence=ok,
        sentences=[too_long],
        truncate_input_tokens=truncate_input_tokens,
    )

    loaded_model.run_sentence_similarities(
        source_sentences=[too_long],
        sentences=[ok],
        truncate_input_tokens=truncate_input_tokens,
    )
    loaded_model.run_sentence_similarities(
        source_sentences=[ok],
        sentences=[too_long],
        truncate_input_tokens=truncate_input_tokens,
    )

    # reranker test both query and document text
    loaded_model.run_rerank_query(
        query=too_long,
        documents=[{"text": ok}],
        truncate_input_tokens=truncate_input_tokens,
    )
    loaded_model.run_rerank_query(
        query=ok,
        documents=[{"text": too_long}],
        truncate_input_tokens=truncate_input_tokens,
    )

    loaded_model.run_rerank_queries(
        queries=[too_long],
        documents=[{"text": ok}],
        truncate_input_tokens=truncate_input_tokens,
    )
    loaded_model.run_rerank_queries(
        queries=[ok],
        documents=[{"text": too_long}],
        truncate_input_tokens=truncate_input_tokens,
    )


@pytest.mark.parametrize(
    "truncate_input_tokens", [1, 2, 3, 4, 99, 100, 101, 510, 511, 512, -1]
)
def test_embeddings_with_truncation(truncate_input_tokens, loaded_model):
    """verify that results are as expected with truncation"""

    max_len = loaded_model.model.max_seq_length - 2
    if truncate_input_tokens is None or truncate_input_tokens < 0:
        # For -1 we don't truncate, but sentence-transformers will truncate at max_seq_length - 2
        repeat = max_len
    else:
        repeat = min(
            truncate_input_tokens, max_len
        )  # max_len is used when we need -2 for begin/end

    # Build a text like "x x x.. x " with room for one more token
    repeat = repeat - 1  # space for the final x or y token to show difference

    base = ""
    if repeat > 0:
        base = "x " * repeat  # A bunch of "x" tokens
    x = base + "x"  # One last "x" that will not get truncated
    y = base + "y"  # A different last character "y" not truncated
    z = y + "z"  # Add token "z" after "y". This should get truncated.

    res = loaded_model.run_embeddings(
        texts=[base, x, y, z], truncate_input_tokens=truncate_input_tokens
    )
    vectors = res.results.vectors  # vectors from batch embeddings

    # Compare with results from individual embedding calls in a loop
    loop_res = []
    for t in [base, x, y, z]:
        r = loaded_model.run_embedding(
            text=t, truncate_input_tokens=truncate_input_tokens
        )
        loop_res.append(r)
    loop_vectors = [
        r.result for r in loop_res
    ]  # vectors from loop of single embedding calls

    assert len(vectors) == len(loop_vectors), "expected the same length vectors"
    # compare the vectors from batch with the single calls
    for i, e in enumerate(vectors):
        assert np.allclose(e.data.values, loop_vectors[i].data.values)

    # x...xyz is the same as x...xy because that is exactly where truncation worked
    assert len(vectors[2].data.values) == len(vectors[3].data.values)
    assert np.allclose(vectors[2].data.values, vectors[3].data.values)
    for i in range(len(vectors[2].data.values)):
        assert approx(vectors[2].data.values[i]) == approx(vectors[3].data.values[i])

    # Make sure the base, x, y are not a match (we kept the significant last char)
    assert not np.allclose(vectors[0].data.values, vectors[1].data.values)
    assert not np.allclose(vectors[0].data.values, vectors[2].data.values)
    assert not np.allclose(vectors[1].data.values, vectors[2].data.values)


def test__with_retry_happy_path(loaded_model):
    """works with args/kwargs, no problems"""
    loaded_model._with_retry(print, "hello", "world", sep="<:)>", end="!!!\n")


def test__with_retry_fail(loaded_model):
    """fn never works, loops then raises the exception"""

    def fn():
        raise (ValueError("always fails with ValueError"))

    with pytest.raises(ValueError):
        loaded_model._with_retry(fn)


def test__with_retry_fail_fail(loaded_model, monkeypatch):
    """fn needs a few tries, tries twice and fails."""

    monkeypatch.setattr(loaded_model, "RETRY_COUNT", 1)  # less than 3 tries

    def generate_ints():
        yield from range(9)  # More than enough for retry loop

    ints = generate_ints()

    def fail_fail_win():
        for i in ints:
            if i < 2:  # fail, fail
                raise (ValueError(f"fail {i}"))
            else:  # win and return 3
                return i + 1

    # Without a third try raises first exception
    with pytest.raises(ValueError) as e:
        loaded_model._with_retry(fail_fail_win)

    assert e.value.args[0] == "fail 0", "expected first exception 'fail 0'"


def test__with_retry_fail_fail_win(loaded_model, monkeypatch):
    """fn needs a few tries, logs, loops and succeeds"""

    monkeypatch.setattr(loaded_model, "RETRY_COUNT", 6)  # test needs at least 3 tries

    def generate_ints():
        yield from range(9)  # More than enough for retry loop

    ints = generate_ints()

    def fail_fail_win():
        for i in ints:
            if i < 2:  # fail, fail
                raise (ValueError("fail, fail"))
            else:  # win and return 3
                return i + 1

    # Third try did not raise an exception. Returns 3.
    assert 3 == loaded_model._with_retry(fail_fail_win)


def test_env_val_to_bool():
    assert not utils.env_val_to_bool(None)
    assert not utils.env_val_to_bool("")
    assert not utils.env_val_to_bool("   ")
    assert not utils.env_val_to_bool(0)
    assert not utils.env_val_to_bool("0")
    assert not utils.env_val_to_bool(" False ")
    assert not utils.env_val_to_bool("  false   ")
    assert not utils.env_val_to_bool("   fAlSE    ")

    assert utils.env_val_to_bool(1)
    assert utils.env_val_to_bool("1")
    assert utils.env_val_to_bool(" True ")
    assert utils.env_val_to_bool("  true   ")
    assert utils.env_val_to_bool("   tRuE    ")


def test_env_val_to_int():
    expected_default = 12345
    assert expected_default == utils.env_val_to_int(None, expected_default)
    assert expected_default == utils.env_val_to_int("", expected_default)
    assert expected_default == utils.env_val_to_int("   ", expected_default)
    assert expected_default == utils.env_val_to_int(" ss ", expected_default)
    assert expected_default == utils.env_val_to_int("  sss   ", expected_default)
    assert expected_default == utils.env_val_to_int("   ssss    ", expected_default)

    assert 0 == utils.env_val_to_int(0, expected_default)
    assert 0 == utils.env_val_to_int("0", expected_default)
    assert 0 == utils.env_val_to_int(False, expected_default)
    assert 456 == utils.env_val_to_int("456", expected_default)
    assert 456 == utils.env_val_to_int(" 456 ", expected_default)
    assert 1 == utils.env_val_to_int(True, expected_default)


@pytest.mark.parametrize(
    # `expected_count` are valid for the `BertForSequenceClassification` model.
    ["texts", "expected_count"],
    [
        # Only tokens requiring model attention is counted.
        # [PAD] doesn't attract model attention, but [CLS] and [SEP] does
        # [CLS] 5 normal tokens [SEP]
        (["12345"], 5 + 2),
        # [CLS] 5 normal [SEP], [CLS] 4 normal [SEP] [PAD]
        (["12 345", "6 789"], 9 + 4),
    ],
)
def test_sum_token_count_no_truncation(texts, expected_count, loaded_model):

    tokenized = loaded_model.model.tokenizer(
        texts,
        return_attention_mask=True,
        return_token_type_ids=False,
        return_overflowing_tokens=True,
        return_offsets_mapping=True,
        return_length=True,
        return_tensors="pt",
        truncation=True,
        padding=True,
        max_length=loaded_model.model.max_seq_length,
    )
    token_count = sum_token_count(
        tokenized,
        truncate_only=False,
    )

    assert token_count == expected_count


@pytest.mark.parametrize(
    # `expected_count` are valid for the `BertForSequenceClassification` model.
    ["texts", "truncate", "expected_count"],
    [
        # Only tokens requiring model attention is counted.
        # [PAD] doesn't attract model attention, but [CLS] and [SEP] does
        #
        # All encodings: [CLS] 12345 [SEP]
        # No truncation
        (["12345"], 10, 7),
        # All encodings: [CLS] 123 [SEP] + [CLS] 45 [SEP] [PAD]
        # Only truncated: [CLS] 123 [SEP]
        (["12345"], 5, 3 + 2),
        #
        # All encodings: [CLS] 123 [SEP] + [CLS] 45 [SEP] [PAD], [CLS] 678 [SEP] + [CLS] 9 [SEP] [PAD] [PAD]
        # Only truncated: [CLS] 123 [SEP] , [CLS] 678 [SEP]
        (["12 345", "6 789"], 5, (3 + 2) + (3 + 2)),
    ],
)
def test_sum_token_count_with_truncation(texts, truncate, expected_count, loaded_model):
    tokenized = loaded_model.model.tokenizer(
        texts,
        return_attention_mask=True,
        return_token_type_ids=False,
        return_overflowing_tokens=True,
        return_offsets_mapping=True,
        return_length=True,
        return_tensors="pt",
        truncation=True,
        padding=True,
        max_length=truncate,
    )
    token_count = sum_token_count(
        tokenized,
        truncate_only=True,
    )

    assert token_count == expected_count


@pytest.mark.parametrize(
    "truncate_input_tokens", [0, 1, 2, 3, 4, 99, 100, 101, 510, 511, 512, 513, -1]
)
def test_encoding_order(loaded_model: EmbeddingModule, truncate_input_tokens):
    """Confirm that encoding doesn't modify the original sort order"""
    separate_embeddings = [
        loaded_model.run_embedding(text=i, truncate_input_tokens=truncate_input_tokens)
        for i in MANY_INPUTS
    ]
    combined_embeddings = loaded_model.run_embeddings(
        texts=MANY_INPUTS, truncate_input_tokens=truncate_input_tokens
    )

    separate_vectors = [
        e.to_dict()["result"]["data"]["values"] for e in separate_embeddings
    ]
    combined_vectors = [
        e["data"]["values"] for e in combined_embeddings.to_dict()["results"]["vectors"]
    ]

    assert len(separate_vectors) == len(
        combined_vectors
    ), "expected the same number separate and combined embeddings"

    # test order by comparing value of individual embeddings in sequence
    for i, e in enumerate(separate_vectors):
        assert np.allclose(e, combined_vectors[i], rtol=1e-03, atol=1e-05)

    # test expected failure case by reordering
    shifted_separate_vectors = separate_vectors[1:] + [separate_vectors[0]]

    for i, e in enumerate(shifted_separate_vectors):
        assert e != separate_vectors[i], "expected order to be have been altered"
        assert (
            not approx(e) == combined_vectors[i]
        ), "expected altered order to not match combined vectors"
        assert not np.allclose(
            e, combined_vectors[i], rtol=1e-05, atol=1e-08
        ), "expected altered order to not match combined"


@pytest.mark.parametrize(
    ("mapping", "expected"),
    [
        ([0, 0, 0, 0, 0], [0]),
        ([0, 1, 2, 3, 4], [0, 1, 2, 3, 4]),
        ([0, 0, 1, 1, 1, 2], [0, 2, 5]),
        ([], []),
    ],
)
def test_get_sample_start_indexes(mapping, expected):
    mock_tokenized = {
        "overflow_to_sample_mapping": torch.Tensor(mapping).type(torch.int8)
    }
    assert get_sample_start_indexes(mock_tokenized) == expected


def test_encode_extensions(loaded_model):
    # loaded model can return_token_count
    ret = loaded_model._encode_with_retry("text here", return_token_count=True)
    assert isinstance(ret, Tuple)
    assert isinstance(ret[0], np.ndarray)
    assert isinstance(ret[1], int)
    ret = loaded_model._encode_with_retry("text here", return_token_count=False)
    assert isinstance(ret, np.ndarray)

    # Make sure use with un-wrapped SentenceTransformer model is unaffected by extended params or return tokens
    ret = BOOTSTRAPPED_MODEL._encode_with_retry(
        "text here",
        return_token_count=True,
        truncate_input_tokens=123,
        implicit_truncation_errors=False,
    )
    assert isinstance(ret, np.ndarray)
    BOOTSTRAPPED_MODEL._encode_with_retry(
        "text here"
    )  # and no KeyError trying to remove non-existing keys


@pytest.mark.parametrize(
    "truncate_input_tokens",
    [0, 1, 2, 3, 4, 5, 99, 100, 101, 300, 510, 511, 512, 513, 1000, -1],
)
def test_same_same(loaded_model: EmbeddingModule, truncate_input_tokens):
    """Confirm that same text gives same results"""

    inputs = ["What is generative ai?", "What is generative ai?", "different"]

    # First ensuring that batch input vs loop over inputs is the same
    separate_embeddings = [
        loaded_model.run_embedding(text=i, truncate_input_tokens=truncate_input_tokens)
        for i in inputs
    ]
    combined_embeddings = loaded_model.run_embeddings(
        texts=inputs, truncate_input_tokens=truncate_input_tokens
    )

    separate_vectors = [
        e.to_dict()["result"]["data"]["values"] for e in separate_embeddings
    ]
    combined_vectors = [
        e["data"]["values"] for e in combined_embeddings.to_dict()["results"]["vectors"]
    ]

    assert len(separate_vectors) == len(
        combined_vectors
    ), "expected the same number separate and combined embeddings"

    # test order by comparing value of individual embeddings in sequence
    for i, e in enumerate(separate_vectors):
        assert np.allclose(e, combined_vectors[i])

    # Next ensuring that the two identical sentences yield identical results (and 3rd does not)
    assert np.allclose(combined_vectors[0], combined_vectors[1], rtol=1e-05, atol=1e-08)
    assert not np.array_equal(combined_vectors[1], combined_vectors[2])
    assert np.allclose(separate_vectors[0], separate_vectors[1], rtol=1e-05, atol=1e-08)
    assert not np.allclose(
        separate_vectors[1], separate_vectors[2], rtol=1e-05, atol=1e-08
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This toolkit utility contains functions to operate on torch distribution

NOTE: Content of this file are heavily influenced by torch/distributed/run.py

Ref: https://github.com/pytorch/pytorch/blob/main/torch/distributed/run.py
"""

# Standard
import os

# Third Party
from torch import cuda
from torch.distributed.launcher.api import LaunchConfig
import torch.distributed as dist

# First Party
import alog

log = alog.use_channel("TRCH_RN")


def initialize_torch_distribution(world_size, rank=0, backend="gloo|nccl"):

    if dist.is_available():
        log.debug(
            "Initializing process group - backend %s, rank %d, world size %d",
            backend,
            rank,
            world_size,
        )
        dist.init_process_group(backend=backend, world_size=world_size, rank=rank)


def determine_local_world_size():
    """Function to automatically deduce the world size based on
    available processors.

    NOTE: This function will try to use ALL gpus accessible to it
    """

    if cuda.is_available():
        num_proc = cuda.device_count()
        log.info("Cuda devices available! Using %d devices.", num_proc)
        return num_proc
    # Fall back to using the OS cpu count
    # TODO: Callibrate this to some reasonable default...
    num_proc = os.cpu_count()
    log.info("Cuda devices NOT available! Using CPU %d processes.", num_proc)
    return num_proc


def get_torch_elastic_launch_config(
    master_addr: str,
    master_port: str,
    start_method: str = "spawn",
    max_restarts=3,
) -> LaunchConfig:

    # Constants; we assume everything executes on the same node
    min_nodes = 1
    max_nodes = 1
    rdzv_configs = {"rank": 0}

    nproc_per_node = determine_local_world_size()

    if "OMP_NUM_THREADS" not in os.environ and nproc_per_node > 1:
        omp_num_threads = 1
        log.warning(
            "\n*****************************************\n"
            "Setting OMP_NUM_THREADS environment variable for each process to be "
            "%s in default, to avoid your system being overloaded, "
            "please further tune the variable for optimal performance in "
            "your application as needed. \n"
            "*****************************************",
            omp_num_threads,
        )
        # This env variable will be passed down to the subprocesses
        os.environ["OMP_NUM_THREADS"] = str(omp_num_threads)

    return LaunchConfig(
        min_nodes=min_nodes,
        max_nodes=max_nodes,
        nproc_per_node=nproc_per_node,
        start_method=start_method,
        rdzv_backend="static",
        rdzv_endpoint=f"{master_addr}:{master_port}",
        rdzv_configs=rdzv_configs,
        max_restarts=max_restarts,
    )



# Utils imports should always come first, because they ensure caikit_nlp
# is added to the syspath if not running inside of a container.
# Standard
import json
import time

# Third Party
from datasets import load_dataset
from utils import SUPPORTED_DATASETS, load_model
import torch

# First Party
import caikit

# Local
from caikit_nlp import data_model
import caikit_nlp

NUM_SAMPLES_TO_RUN = 100

model_path = "prompt_prefixes/sample_prompt"
# Grab the test stream from the twitter dataset
test_stream = SUPPORTED_DATASETS["twitter_complaints"].dataset_loader()[2]

# Load the local block
local_model = load_model(is_distributed=False, model_path=model_path)
# Load the TGIS backed block; this will kill TGIS if a container is already running, then restart.
distributed_model = load_model(is_distributed=True, model_path=model_path)

preds = []
for datum in test_stream:
    dis_res = distributed_model.run(datum.input)
    local_res = local_model.run(datum.input)
    preds.append(
        {
            "input": datum.input,
            "output": datum.output,
            "local_prediction": local_res.text.split(":")[-1].strip(),
            "distributed_prediction": dis_res.text.split(":")[-1].strip(),
        }
    )
    if len(preds) >= NUM_SAMPLES_TO_RUN:
        break


with open("preds.json", "w") as f:
    json.dump(preds, f, sort_keys=True, indent=4)

num_matching = 0
num_local_correct = 0
num_distributed_correct = 0
num_mismatching = 0
for x in preds:
    if x["output"] == x["local_prediction"]:
        num_local_correct += 1

    if x["output"] == x["distributed_prediction"]:
        num_distributed_correct += 1

    if x["local_prediction"] == x["distributed_prediction"]:
        num_matching += 1
    else:
        num_mismatching += 1

print("----- Metrics -----")
print("Num correct [local block via PEFT]: {}".format(num_local_correct))
print("Num correct [distributed via TGIS]: {}".format(num_distributed_correct))
print("Num matching remote / local preds: {}".format(num_matching))
print("Num not matching remote / local preds: {}".format(num_mismatching))



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Huggingface auto causal LM resource type
"""
# Standard
from collections.abc import Mapping
from typing import List, Union

# Third Party
from transformers import (
    AutoModelForCausalLM,
    BatchEncoding,
    DataCollatorForLanguageModeling,
)
from transformers.models.auto import modeling_auto
import torch

# First Party
from caikit.core.data_model import DataStream
from caikit.core.exceptions import error_handler
from caikit.core.modules import module
import alog

# Local
from ...data_model import GenerationTrainRecord, PromptOutputModelType
from ...toolkit.verbalizer_utils import render_verbalizer
from .base import PretrainedModelBase

log = alog.use_channel("HFRCLM")
error = error_handler.get(log)


@module(
    id="6759e891-287b-405b-bd8b-54a4a4d51c24",
    name="HF Transformers Auto Causal LM",
    version="0.1.0",
)
class HFAutoCausalLM(PretrainedModelBase):
    """This resource (module) wraps a handle to a Huggingface AutoModelForCausalLM"""

    MODEL_TYPE = AutoModelForCausalLM
    SUPPORTED_MODEL_TYPES = modeling_auto.MODEL_FOR_CAUSAL_LM_MAPPING_NAMES
    TASK_TYPE = "CAUSAL_LM"
    PROMPT_OUTPUT_TYPES = [PromptOutputModelType.DECODER]
    MAX_NUM_TRANSFORMERS = 1
    REQUIRES_TOKEN_UNWRAPPING = True

    @classmethod
    def tokenize_function(
        cls,
        example: Union[GenerationTrainRecord, Mapping],
        tokenizer: "AutoTokenizer",
        max_source_length: int,
        max_target_length: int,
        verbalizer: Union[None, str] = None,
        task_ids: Union[None, int] = None,
        use_seq2seq_approach: bool = True,
        chunk_size: int = 128,
        drop_remainder: bool = False,
    ) -> Union[DataStream[BatchEncoding], BatchEncoding]:
        """Tokenization function to be used for causallm training; this function consumes a
        GenerationTrainRecord object and applies the verbalizer to it followed by
        the model tokenizer. Due to the nature of our training data with src/target seqs,
        each sample yields one example per token in the target sequence.

        Args:
            example: GenerationTrainRecord | Mapping
                Training data model object to convert a form we can learn on, or a Mapping
                that has keys input/output.
            tokenizer: AutoTokenizer
                Tokenizer object to be applied to input records.
            max_source_length: int
                Maximum length for input sequences.
            max_target_length: int
                Maximum length for output sequences.
            verbalizer: Union[None, str]
                Verbalizer to be rendered into each text.
            task_ids: Union[None, int]
                Task IDs to be used for multiprompt tuning.
            use_seq2seq_approach: bool
                Indicates whether or not we should use a sequence style approach
                or use chunking parameters.
            chunk_size: int
                unsigned int value to be used for chunk size.
                Only used if use_seq2seq_approach=True.
            drop_remainder: bool
                Whether or not to keep the residual as an extra chunk if the
                total number of tokens is not divisible by the chunk size.
                Only used if use_seq2seq_approach=True.

        Returns:
            Union[DataStream[BatchEncoding], BatchEncoding]
                stream of encoded tokenization output corresponding to the input example
                or a single batch encoding object containing 1+ tokenized results.
        """
        ### Things common to all Causal LM tokenization approaches
        # Extract the source & target from our provided inputs
        source, target = cls.decompose_example_io(example)
        # Determine if our mapped inputs are in batched mode or not
        batched_mode = isinstance(source, list) and isinstance(target, list)

        # TODO: Handle batched verbalizer stuff!
        if batched_mode and verbalizer is not None:
            raise NotImplementedError(
                "Verbalizer rendering not implemented for batch mode"
            )
        source = (
            source if verbalizer is None else render_verbalizer(verbalizer, example)
        )
        # Treat this as a seq2seq type problem. Note that this implementation is different
        # from the seq2seq tokenization function even though it is conceptually similar due
        # to sequence length / padding requirements assumed internally by causal LMs.
        if use_seq2seq_approach:
            return cls._causal_lm_padding_as_seq2seq(
                tokenizer=tokenizer,
                source=source,
                target=target,
                max_source_length=max_source_length,
                max_target_length=max_target_length,
                task_ids=task_ids,
            )
        # Do causal language model chunking
        return cls._causal_lm_as_chunked(
            tokenizer=tokenizer,
            source=source,
            target=target,
            max_source_length=max_source_length,
            max_target_length=max_target_length,
            batched_mode=batched_mode,
            task_ids=task_ids,
            chunk_size=chunk_size,
            drop_remainder=drop_remainder,
        )

    def _get_data_collator(self, **kwargs) -> "transformers.DataCollator":
        """Function to return appropriate data collator based on resource.

        DataCollatorForLanguageModeling is used here which will dynamically
        padded to maximum length of a batch if they are not all of the same
        length.

        NOTE: If mlm (masked language modeling) is not passed in kwargs,
        this function will automatically set it to `False`.

        FIXME: This should be consolidated with what is in the prompt tuning
        module, which currently does its own collator management outside of the
        resource classes.

        Args:
            **kwargs:
                All the keyword arguments passed to this function
                will get filtered out to appropriate ones that are
                applicable to implemented data collator.
        Returns:
            transformers.DataCollator
                Collator to be used for causal language modeling.
        """

        applicable_args = ["mlm", "pad_to_multiple_of"]
        collator_kwargs = {key: kwargs[key] for key in applicable_args if key in kwargs}

        if "mlm" not in collator_kwargs:
            collator_kwargs["mlm"] = False

        return DataCollatorForLanguageModeling(
            tokenizer=self._tokenizer, return_tensors="pt", **collator_kwargs
        )

    ### Tokenization strategy implementations
    # Chunked causal language modeling
    @classmethod
    def _causal_lm_as_chunked(
        cls,
        tokenizer: "AutoTokenizer",
        source: str,
        target: str,
        max_source_length: int,
        max_target_length: int,
        batched_mode: bool,
        task_ids: Union[None, int],
        chunk_size: int,
        drop_remainder: bool,
    ) -> Union[DataStream[BatchEncoding], BatchEncoding]:
        """Given a source and target string, build the chunked concatenated sequence and formulate
        the batch encoded chunks for the sequence. If running in batch mode, the chunks will be
        collapsed into a single batch encoding for the whole sequence. Otherwise, each chunk will
        placed in its own BatchEncoding and encapsulated within a datastream.

        Args:
            tokenizer: AutoTokenizer
                Tokenizer object to be applied to input records.
            source: str
                Raw source string.
            target: str
                Raw target string.
            max_source_length: int
                Maximum length for input sequences.
            max_target_length: int
                Maximum length for output sequences.
            batched_mode: bool
                Whether or not we should produce a stream of encodings or a single
                encoding representing all of the chunked sequence.
            task_ids: Union[None, int]
                Task IDs to be used for multiprompt tuning.
            chunk_size: int
                unsigned int value to be used for chunk size.
            drop_remainder: bool
                Whether or not to keep the residual as an extra chunk if the
                total number of tokens is not divisible by the chunk size.

        Returns:
            Union[DataStream[BatchEncoding], BatchEncoding]
                Encoded chunked sequence as a stream or batch encoding object.
        """
        source_ids = tokenizer(source, max_length=max_source_length, truncation=True)
        target_ids = tokenizer(target, max_length=max_target_length, truncation=True)

        # Force everything to a list of batch encodings; for non-batch mode, this just
        # puts it into a list. For batch mode, we get a list of batch encodings,
        # allowing us to standardize subsequent processing a bit.
        #
        # For example, given chunk size 2, we might have something like:
        # [
        #   {'input_ids': [31, 48], 'attention_mask': [1, 1]},
        #   {'input_ids': [47, 1], 'attention_mask': [1, 1]},
        #   ...
        # ]
        # (where the above objects are batch encodings, which are a subclass of dict)
        source_id_chunks = cls._force_to_batch_encoding_list_of_chunks(
            source_ids, target_ids, batched_mode, task_ids, chunk_size, drop_remainder
        )

        def generator_func():
            for chunk in source_id_chunks:
                yield chunk

        chunk_stream = DataStream(generator_func)
        # If it's batch mode, collapse down into one encoding batch object
        if batched_mode:
            return cls._collapse_stream_into_encoding(chunk_stream)
        # Otherwise just produce the stream to be chained
        # NOTE: it might be a good idea to deprecate this to force standardization
        # onto using batch encodings the way that they are intended to be
        return chunk_stream

    @staticmethod
    def _force_to_batch_encoding_list_of_chunks(
        source_ids: BatchEncoding,
        target_ids: BatchEncoding,
        batch_mode: bool,
        task_ids: Union[None, int],
        chunk_size: int,
        drop_remainder: bool,
    ) -> List[BatchEncoding]:
        """Forces our inputs into either a single batch encoding (if we aren't running in batch
        mode), or a list of Batch Encodings. I.e., a list of dicts instead of a dict of lists.
        The primary reason that we do this is to allow us to easily map a common generator
        func, regardless of whether or not we're running in batched mode.

        Args:
            source_ids: BatchEncoding
                Source ID batch encoding; target ID info will be merged into this one.
            target_ids: BatchEncoding
                Target ID batch encoding; information will be merged into source ID encoding.
            batch_mode: bool
                Whether or not we are processing a batch.
            task_ids: Union[None, int]
                Optional task IDs for MPT to be propagated to produced encodings.
            chunk_size: int
                unsigned int value to be used for chunk size.
            drop_remainder: bool
                Whether or not to keep the residual as an extra chunk if the
                total number of tokens is not divisible by the chunk size.

        Returns:
            List[BatchEncoding]
                List of batch encodings, each of which encapsulates the contents
                of a single chunk.
        """
        if not batch_mode:
            HFAutoCausalLM._concatenate_encodings(source_ids, target_ids)
            chunks = HFAutoCausalLM._split_encoding_into_chunks(
                encoding=source_ids,
                chunk_size=chunk_size,
                drop_remainder=drop_remainder,
                task_ids=task_ids,
            )
            return chunks
        # Otherwise we need to expand the dict along its keys,
        # mapping all of its encapsulated objects to new items.
        encodings = []
        id_keys = source_ids.keys()
        key = None
        error.value_check(
            "<NLP94411004E>",
            source_ids.keys(),
            "Source ID batch encoding must have keys",
        )

        for batch_idx in range(len(source_ids.input_ids)):
            new_encoding = BatchEncoding()
            for key in id_keys:
                new_encoding[key] = (
                    source_ids[key][batch_idx] + target_ids[key][batch_idx]
                )
            chunks = HFAutoCausalLM._split_encoding_into_chunks(
                encoding=new_encoding,
                chunk_size=chunk_size,
                drop_remainder=drop_remainder,
                task_ids=task_ids,
            )
            # Chunks are held as a list of lists
            encodings += chunks
        return encodings

    @staticmethod
    def _concatenate_encodings(left: BatchEncoding, right: BatchEncoding) -> None:
        """Given two batch encodings, combine their entries into a single encoding.

        Args:
            left: BatchEncoding
                Encoding representing left sequence, which will be updated in place.
                Corresponds to source.
            right: BatchEncoding
                Encoding representing right sequence, which will be stacked onto the left
                encoding. Corresponds to target.
        """
        for k in left.keys():
            left[k].extend(right[k])

    @staticmethod
    def _split_encoding_into_chunks(
        encoding: BatchEncoding,
        chunk_size: int,
        drop_remainder: bool,
        task_ids: Union[None, int],
    ) -> List[BatchEncoding]:
        """Fetch the chunked batch encoding objects from the concatenated encoding.

        Args:
            encoding: BatchEncoding
                BatchEncoding holding the concatenated source/target for one example.
            chunk_size: int
                unsigned int value to be used for chunk size.
            drop_remainder: bool
                Whether or not to keep the residual as an extra chunk if the
                total number of tokens is not divisible by the chunk size.
            task_ids: Union[None, int]
                Optional task IDs for MPT to be propagated to produced encodings.

        Returns:
            List[BatchEncoding]
                List of encodings, where each encoding represents one chunk.
        """
        chunked_encodings = []
        # all encoding keys have the same length list values; we just use input ids
        tok_len = len(encoding["input_ids"])
        # Build a batch encoding for every chunk; for each data,
        # use the slice for all keys inside of the source_encoding.
        if tok_len >= chunk_size:
            slice_len = (tok_len // chunk_size) * chunk_size
            # If we have a remainder and we don't want to drop it, add a new chunk
            if not drop_remainder and slice_len != tok_len:
                slice_len += chunk_size
        # We just have one big chunk
        else:
            slice_len = tok_len
        chunked_encodings = [
            BatchEncoding(
                data={
                    k: v[chunk_num : chunk_num + chunk_size]
                    for k, v in encoding.items()
                }
            )
            for chunk_num in range(0, slice_len, chunk_size)
        ]
        for enc in chunked_encodings:
            enc["task_ids"] = task_ids
        return chunked_encodings

    @staticmethod
    def _collapse_stream_into_encoding(
        stream: DataStream[BatchEncoding],
    ) -> BatchEncoding:
        """Given a stream batch encodings, collapse them back into
        one encoding, i.e., the return value of the batch encoding.

        Args:
            streams: List[DataStream[BatchEncoding]]
                Objects to be collapsed into a single encoding.
            encoding_keys: dict_keys
                Dictionary keys to copy over from batch encoding list.

        Returns:
            BatchEncoding
                Collapsed batch encoding to be returned from tokenization func.
        """
        encoding_keys = None
        new_encoding = BatchEncoding()
        # Now build the individual lists lists for each entry
        for enc in stream:
            # Initialize the existing keys in the new encoding
            if encoding_keys is None:
                encoding_keys = enc.keys()
                for k in encoding_keys:
                    new_encoding[k] = []
            for k in encoding_keys:
                new_encoding[k].append(enc[k])
        return new_encoding

    # Causal language modeling as a sequence to sequence problem
    @staticmethod
    def _causal_lm_padding_as_seq2seq(
        tokenizer: "AutoTokenizer",
        source: str,
        target: str,
        max_source_length: int,
        max_target_length: int,
        task_ids: Union[None, int],
    ) -> BatchEncoding:
        """Tokenize the example as a seq2seq type problem; this is conceptually similar to
        what seq2seq tokenization is doing, but some care needs be taken to ensure the labels
        are the same length as the input sequence because of the shifting mechanism implemented
        in most causal language models.

        Collator compatability is extremely important here; because we are setting the labels
        directly, we should NOT use the causal lm collator, otherwise it will clobber it with a
        shifted input sequence.

        Args:
            tokenizer: AutoTokenizer
                Tokenizer object to be applied to input records.
            source: str
                Raw source string.
            target: str
                Raw target string.
            max_source_length: int
                Maximum length for input sequences.
            max_target_length: int
                Maximum length for output sequences.
            task_ids: Union[None, int]
                Optional task IDs for MPT to be propagated to produced encodings.
        Returns:
            BatchEncoding
                BatchEncoding object corresponding to this example, where the input_ids,
                attention_mask, and labels all have the same length, i.e.,
                [max_source_length + max_target_length + 1].
        """
        IGNORE_ID = -100
        # ID of the token to append after our target string; this should generally be pad / EOS
        FINAL_TOK_ID = tokenizer.eos_token_id
        max_concat_length = max_source_length + max_target_length + 1

        # Truncate based on max source or max target length before considering as a joined sequence
        model_inputs = tokenizer(source, truncation=True, max_length=max_source_length)
        labels = tokenizer(target, truncation=True, max_length=max_target_length + 1)

        # Combine the source + target strings into the source input IDs
        # This makes the source and target the same length, and then masks the source out of the
        # target IDs, and updates the length of the attention vector to be evenly spread on the
        # whole combined sequence
        sample_input_ids = model_inputs["input_ids"]
        label_input_ids = labels["input_ids"] + [FINAL_TOK_ID]
        model_inputs["input_ids"] = sample_input_ids + label_input_ids
        labels["input_ids"] = [IGNORE_ID] * len(sample_input_ids) + label_input_ids
        model_inputs["attention_mask"] = [1] * len(model_inputs["input_ids"])
        # Now we have to update everything to be the max length of the tokenizer, then pad &
        # ensure all of the padded stuff we have added has attention weights of 0.
        sample_input_ids = model_inputs[
            "input_ids"
        ]  # NOTE - combined source + target + <FINAL_TOK_ID>

        label_input_ids = labels["input_ids"]
        model_inputs = tokenizer.pad(
            model_inputs, padding="max_length", max_length=max_concat_length
        )

        if tokenizer.padding_side.lower() == "left":
            labels["input_ids"] = [IGNORE_ID] * (
                max_concat_length - len(sample_input_ids)
            ) + label_input_ids
        else:
            labels["input_ids"] = label_input_ids + [IGNORE_ID] * (
                max_concat_length - len(sample_input_ids)
            )

        model_inputs["input_ids"] = torch.tensor(
            model_inputs["input_ids"][:max_concat_length]
        )
        model_inputs["attention_mask"] = torch.tensor(
            model_inputs["attention_mask"][:max_concat_length]
        )

        labels["input_ids"] = torch.tensor(labels["input_ids"][:max_concat_length])
        model_inputs["labels"] = labels["input_ids"]
        model_inputs["task_ids"] = task_ids
        return model_inputs



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the TGISAutoFinder"""

# Standard
from contextlib import contextmanager
from typing import Optional
from unittest.mock import patch
import copy

# Third Party
import pytest

# First Party
from caikit.config.config import merge_configs
from caikit.core.model_manager import ModelManager
from caikit_tgis_backend import TGISBackend
import aconfig
import caikit

# Local
from caikit_nlp.model_management import tgis_auto_finder
from caikit_nlp.modules.text_generation.text_generation_tgis import TextGenerationTGIS

## Helpers #####################################################################

# Convenient aliases
LOCAL_INITIALIZER_NAME = tgis_auto_finder.TGISAutoFinder._LOCAL_INITIALIZER_NAME_KEY
TGIS_BACKEND_PRIORITY = tgis_auto_finder.TGISAutoFinder._TGIS_BACKEND_PRIORITY_KEY


def make_tgis_config(hostname: str):
    return {
        "type": TGISBackend.backend_type,
        "config": {
            "test_connections": False,
            "connection": {
                "hostname": hostname,
            },
        },
    }


@contextmanager
def temp_model_manager(
    auto_finder_config: Optional[dict] = None,
    backend_priority: Optional[list] = None,
    local_initializer_key: str = "default",
):
    global_config = copy.deepcopy(getattr(caikit.config.config, "_CONFIG"))
    if backend_priority is None:
        backend_priority = [make_tgis_config("foo.bar:123")]
    updated_config = merge_configs(
        global_config,
        {
            "model_management": {
                "finders": {
                    "default": {
                        "type": tgis_auto_finder.TGISAutoFinder.name,
                        "config": auto_finder_config or {},
                    },
                },
                "initializers": {
                    local_initializer_key: {
                        "type": "LOCAL",
                        "config": {
                            "backend_priority": backend_priority,
                        },
                    },
                },
            },
        },
    )
    with patch("caikit.core.model_manager.get_config", lambda: updated_config):
        mmgr = ModelManager()
        # NOTE: The TGISAutoFinder relies on searching for the TGISBackend in
        #   the global caikit.core.MODEL_MANAGER and can inadvertently trigger
        #   that global instance to set up an initializer that prefers TGIS over
        #   LOCAL. We need isolation for tests, so we mock that global with this
        #   temporary instance here.
        with patch.object(tgis_auto_finder, "MODEL_MANAGER", new=mmgr):
            mmgr.initialize_components()
            yield mmgr


## Tests #######################################################################


def test_auto_find_tgis_model_ok():
    """Test that a TGIS text-gen model can be auto-found"""
    with temp_model_manager() as mmgr:
        model = mmgr.load("flan-t5-xl")
        assert model
        assert isinstance(model, TextGenerationTGIS)


def test_auto_find_tgis_model_non_default_local_initializer():
    """Test that a TGIS text-gen model can be auto-found when the local
    initializer is not the default
    """
    init_name = "notdefault"
    with temp_model_manager(
        auto_finder_config={LOCAL_INITIALIZER_NAME: init_name},
        local_initializer_key=init_name,
    ) as mmgr:
        model = mmgr.load("flan-t5-xl", initializer=init_name)
        assert model
        assert isinstance(model, TextGenerationTGIS)


def test_auto_find_tgis_model_multiple_tgis_backends_use_first():
    """Test that a TGIS text-gen model can be auto-found when there are
    multiple TGIS backends configured and no explicit priority is given
    """
    with temp_model_manager(
        backend_priority=[
            make_tgis_config("foo.bar:1234"),
            make_tgis_config("baz.bat:4567"),
        ]
    ) as mmgr:
        tgis_be0 = mmgr.get_initializer("default").backends[0]
        tgis_be1 = mmgr.get_initializer("default").backends[1]
        with patch.object(tgis_be0, "get_connection") as get_con_mock0:
            with patch.object(tgis_be1, "get_connection") as get_con_mock1:
                model = mmgr.load("flan-t5-xl")
                assert model
                assert isinstance(model, TextGenerationTGIS)
                get_con_mock0.assert_called()
                get_con_mock1.assert_not_called()


def test_auto_find_tgis_model_multiple_tgis_backends_set_order():
    """Test that a TGIS text-gen model can be auto-found when there are
    multiple TGIS backends configured and priority is explicitly given
    """
    with temp_model_manager(
        backend_priority=[
            make_tgis_config("foo.bar:1234"),
            make_tgis_config("baz.bat:4567"),
        ],
        auto_finder_config={TGIS_BACKEND_PRIORITY: 1},
    ) as mmgr:
        tgis_be0 = mmgr.get_initializer("default").backends[0]
        tgis_be1 = mmgr.get_initializer("default").backends[1]
        with patch.object(tgis_be0, "get_connection") as get_con_mock0:
            with patch.object(tgis_be1, "get_connection") as get_con_mock1:
                model = mmgr.load("flan-t5-xl")
                assert model
                assert isinstance(model, TextGenerationTGIS)
                get_con_mock0.assert_not_called()
                get_con_mock1.assert_called()


def test_bad_config_args():
    """Make sure that all flavors of bad configuration args are handled with
    appropriate errors
    """
    # Bad initializer name type
    with pytest.raises(TypeError):
        tgis_auto_finder.TGISAutoFinder(
            aconfig.Config(
                {LOCAL_INITIALIZER_NAME: 123},
                override_env_vars=False,
            )
        )

    # Bad tgis_backend_priority type
    with pytest.raises(TypeError):
        tgis_auto_finder.TGISAutoFinder(
            aconfig.Config(
                {TGIS_BACKEND_PRIORITY: "1"},
                override_env_vars=False,
            )
        )

    # Invalid tgis_backend_priority index value
    with pytest.raises(ValueError):
        tgis_auto_finder.TGISAutoFinder(
            aconfig.Config(
                {TGIS_BACKEND_PRIORITY: 123},
                override_env_vars=False,
            )
        )

    # Invalid tgis_backend_priority index value
    with pytest.raises(ValueError):
        tgis_auto_finder.TGISAutoFinder(
            aconfig.Config(
                {TGIS_BACKEND_PRIORITY: 123},
                override_env_vars=False,
            )
        )

    # Non-TGIS tgis_backend_priority index value
    with pytest.raises(ValueError):
        tgis_auto_finder.TGISAutoFinder(
            aconfig.Config(
                {TGIS_BACKEND_PRIORITY: 0},
                override_env_vars=False,
            )
        )

    # No TGIS backend available
    with pytest.raises(ValueError):
        tgis_auto_finder.TGISAutoFinder(aconfig.Config({}))


def test_unsupported_tgis_model():
    """Test that attempting to use a TGIS connection for a model that isn't
    supported by the TGIS backend results in not finding the model
    """
    with temp_model_manager(
        backend_priority=[
            {
                "type": TGISBackend.backend_type,
                "config": {
                    "remote_models": {
                        "not-flan-t5-xl": {
                            "hostname": "foo.bar:123",
                        }
                    }
                },
            }
        ],
    ) as mmgr:
        finder = mmgr._finders["default"]
        assert isinstance(finder, tgis_auto_finder.TGISAutoFinder)
        assert finder.find_model("flan-t5-xl") is None


def test_bad_tgis_connection():
    """Test that attempting to use a TGIS connection that can't connect results
    in not finding the model
    """
    with temp_model_manager(
        backend_priority=[
            {
                "type": TGISBackend.backend_type,
                "config": {
                    "remote_models": {
                        "flan-t5-xl": {
                            "hostname": "foo.bar:123",
                        }
                    },
                    "test_connections": True,
                    "connect_timeout": 5,
                },
            }
        ],
    ) as mmgr:
        finder = mmgr._finders["default"]
        assert isinstance(finder, tgis_auto_finder.TGISAutoFinder)
        assert finder.find_model("flan-t5-xl") is None



"""This script illustrates how to fine-tune a model.

Supported model types:
- Seq2Seq LM
"""

# Standard
from typing import Any, Tuple
import argparse
import json
import os
import shutil

# Third Party
from tqdm import tqdm
from transformers import AutoConfig
from utils import (
    ALOG_OPTS,
    SUPPORTED_DATASETS,
    SUPPORTED_METRICS,
    DatasetInfo,
    configure_random_seed_and_logging,
    load_model,
    print_colored,
)

# First Party
from caikit.core.data_model import DataStream
from caikit.core.toolkit import wip_decorator
import alog

# Local
from caikit_nlp.data_model import GenerationTrainRecord, TuningConfig
from caikit_nlp.modules.text_generation import TextGeneration
from caikit_nlp.resources.pretrained_model import (
    HFAutoCausalLM,
    HFAutoSeq2SeqLM,
    PretrainedModelBase,
)

# TODO: Remove me once fine-tuning is out of WIP
wip_decorator.disable_wip()


def get_resource_type(model_name: str) -> PretrainedModelBase:
    """Given a model name, or a path to a model, get the resource type to be initialized.

    Args:
        model_name: str
            Model name or path to the model to be leveraged.

    Returns:
        type
            PretrainedModel subclass wrapping the loaded Transformer model, e.g.,
            a HFAutoCausalLM or HFAutoSeq2SeqLM. We return the type here so that
            we can initialize it later, after we show a nice experiment configuration.
    """
    try:
        model_type = AutoConfig.from_pretrained(model_name).model_type
    except OSError:
        raise ValueError("Failed to load model with name: {}".format(model_name))
    if model_type in HFAutoCausalLM.SUPPORTED_MODEL_TYPES:
        return HFAutoCausalLM
    elif model_type in HFAutoSeq2SeqLM.SUPPORTED_MODEL_TYPES:
        return HFAutoSeq2SeqLM
    raise NotImplementedError(
        "Provided is not supported for any supported resource type!"
    )


### Functions for arg parsing & validation
def parse_args() -> argparse.Namespace:
    """Parse command line arguments.

    Returns:
        argparse.Namespace
            Parsed arguments to be leveraged for fine tuning application.
    """
    parser = argparse.ArgumentParser(
        description="Fine-tuning a text generation model.",
    )
    # Register all of the common args, as well as specific tuning args for subcommands
    register_common_arguments(parser)

    args = parser.parse_args()
    # Reconfigure logging level based on verbosity, while preserving filters etc.

    alog_settings = {**ALOG_OPTS, **{"default_level": "debug"}}
    alog.configure(**alog_settings)
    # Validate common arg values
    validate_common_args(args)
    return args


def register_common_arguments(subparser: argparse.ArgumentParser) -> None:
    """Registers common arguments intended to be shared across all subparsers.

    Args:
        subparser: argparse.ArgumentParser
            Iterable of argument subparsers that should have common args.
    """
    subparser.add_argument(
        "--dataset",
        help="Dataset to use to train prompt vectors. Options: {}".format(
            list(SUPPORTED_DATASETS.keys())
        ),
        default="twitter_complaints",
    )
    subparser.add_argument(
        "--model_name",
        help="Name of base model or path to model to use to train prompt vectors",
        default="t5-small",
    )
    subparser.add_argument(
        "--output_dir",
        help="Name of the directory that we want to export the model to",
        default="sample_tuned_model",
    )
    subparser.add_argument(
        "--num_epochs",
        help="Number of epochs to use for prompt tuning",
        type=int,
        default=10,
    )
    subparser.add_argument(
        "--learning_rate",
        help="Learning rate to use while training",
        type=float,
        default=2e-5,
    ),
    subparser.add_argument(
        "--batch_size", help="Batch size to use while training", type=int, default=8
    )
    subparser.add_argument(
        "--max_source_length",
        help="Maximum source sequence length.",
        default=256,
        type=int,
    )
    subparser.add_argument(
        "--max_target_length",
        help="Maximum target sequence length.",
        default=128,
        type=int,
    )
    subparser.add_argument(
        "--accumulate_steps",
        help="Gradient accumulation steps",
        default=1,
        type=int,
    )
    subparser.add_argument(
        "--evaluate",
        help="Enable evaluation on trained model",
        action="store_true",
    )
    subparser.add_argument(
        "--preds_file",
        help="JSON file to dump raw source / target texts to.",
        default="model_preds.json",
    )
    subparser.add_argument(
        "--torch_dtype",
        help="Torch dtype to use for training",
        type=str,
        default="float16",
    )
    subparser.add_argument(
        "--metrics",
        help="Metrics to calculate. Options: {}".format(list(SUPPORTED_METRICS.keys())),
        nargs="*",
        default=["accuracy"],
    )
    subparser.add_argument(
        "--tgis",
        help="Run inference using TGIS. NOTE: This involves saving and reloading model in TGIS container",
        action="store_true",
    )
    subparser.add_argument(
        "--iterable_dataset",
        help="Enable evaluation on trained model",
        action="store_true",
    )


def validate_common_args(args: argparse.Namespace):
    """Validates common arguments to ensure values make sense; here, we only validate things that
    are not (or should not) be handled within the module.

    Args:
        args: argparse.Namespace
            Parsed args corresponding to one tuning task.
    """
    # Validate that the dataset is one of our allowed values
    if args.dataset not in SUPPORTED_DATASETS:
        raise KeyError(
            "[{}] is not a supported dataset; see --help for options.".format(
                args.dataset
            )
        )
    # Purge our output directory if one already exists.
    if os.path.isdir(args.output_dir):
        print("Existing model directory found; purging it now.")
        shutil.rmtree(args.output_dir)


def show_experiment_configuration(args, dataset_info, model_type) -> None:
    """Show the complete configuration for this experiment, i.e., the model info,
    the resource type we built, the training params, metadata about the dataset where
    possible, and so on.

    Args:
        args: argparse.Namespace
            Parsed args corresponding to one tuning task.
        dataset_info: DatasetInfo
            Dataset information.
        model_type: type
            Resource class corresponding to the base model.
    """
    print_strs = [
        "Experiment Configuration",
        "- Model Name: [{}]".format(args.model_name),
        " |- Inferred Model Resource Type: [{}]".format(model_type),
        "- Dataset: [{}]".format(args.dataset),
        "- Number of Epochs: [{}]".format(args.num_epochs),
        "- Learning Rate: [{}]".format(args.learning_rate),
        "- Batch Size: [{}]".format(args.batch_size),
        "- Output Directory: [{}]".format(args.output_dir),
        "- Maximum source sequence length: [{}]".format(args.max_source_length),
        "- Maximum target sequence length: [{}]".format(args.max_target_length),
        "- Gradient accumulation steps: [{}]".format(args.accumulate_steps),
        "- Enable evaluation: [{}]".format(args.evaluate),
        "- Evaluation metrics: [{}]".format(args.metrics),
        "- Torch dtype to use for training: [{}]".format(args.torch_dtype),
        "- Using iterable dataset: [{}]".format(args.iterable_dataset),
    ]
    # Log and sleep for a few seconds in case people actually want to read this...
    print_colored("\n".join([print_str for print_str in print_strs if print_str]))


def get_model_preds_and_references(
    model, validation_stream, truncate_input_tokens, max_new_tokens
):
    """Given a model & a validation stream, run the model against every example in the validation
    stream and compare the outputs to the target/output sequence.

    Args:
        model
            Fine-tuned Model to be evaluated (may leverage different backends).
        validation_stream: DataStream[GenerationTrainRecord]
            Validation stream with labeled targets that we want to compare to our model's
            predictions.
        truncate_input_tokens: int
            maximum number of tokens to be accepted by the model and rest will be
            truncated.
    Returns:
        Tuple(List)
            Tuple of 2 lists; the model predictions and the expected output sequences.
    """
    model_preds = []
    targets = []

    for datum in tqdm(validation_stream):
        # Local .run() currently prepends the input text to the generated string;
        # Ensure that we're just splitting the first predicted token & beyond.
        raw_model_text = model.run(
            datum.input,
            truncate_input_tokens=truncate_input_tokens,
            max_new_tokens=max_new_tokens,
        ).generated_text
        parse_pred_text = raw_model_text.split(datum.input)[-1].strip()
        model_preds.append(parse_pred_text)
        targets.append(datum.output)
    return (
        model_preds,
        targets,
    )


def export_model_preds(preds_file, predictions, validation_stream):
    """Exports a JSON file containing a list of objects, where every object contains:
        - source: str - Source string used for generation.
        - target: str - Ground truth target label used for generation.
        - predicted_target: str - Predicted model target.

    Args:
        preds_file: str
            Path on disk to JSON file to be written.
        predictions: List
            Model prediction list, where each predicted text excludes source text as a prefix.
        validation_stream: DataStream
            Datastream object of GenerationTrainRecord objects used for validation against a model
            to generate predictions.
    """
    pred_objs = []
    for pred, record in zip(predictions, validation_stream):
        pred_objs.append(
            {
                "source": record.input,
                "target": record.output,
                "predicted_target": pred,
            }
        )
    with open(preds_file, "w") as jfile:
        json.dump(pred_objs, jfile, indent=4, sort_keys=True)


if __name__ == "__main__":
    configure_random_seed_and_logging()
    args = parse_args()
    model_type = get_resource_type(args.model_name)
    # Unpack the dataset dictionary into a loaded dataset & verbalizer
    dataset_info = SUPPORTED_DATASETS[args.dataset]
    show_experiment_configuration(args, dataset_info, model_type)
    # Convert the loaded dataset to a stream
    print_colored("[Loading the dataset...]")
    # TODO - conditionally enable validation stream
    train_stream = dataset_info.dataset_loader()[0]

    # Init the resource & Build the tuning config from our dataset/arg info
    print_colored("[Loading the base model resource...]")
    base_model = model_type.bootstrap(
        args.model_name, tokenizer_name=args.model_name, torch_dtype=args.torch_dtype
    )

    # Then actually train the model & save it
    print_colored("[Starting the training...]")
    model = TextGeneration.train(
        base_model,
        train_stream,
        max_source_length=args.max_source_length,
        max_target_length=args.max_target_length,
        lr=args.learning_rate,
        torch_dtype=args.torch_dtype,
        batch_size=args.batch_size,
        accumulate_steps=args.accumulate_steps,
        num_epochs=args.num_epochs,
        use_iterable_dataset=args.iterable_dataset,
    )

    print_colored("[Training Complete]")

    # Prediction
    sample_text = "summarize: The Inflation Reduction Act lowers prescription drug costs, health care costs, and energy costs. It's the most aggressive action on tackling the climate crisis in American history, which will lift up American workers and create good-paying, union jobs across the country. It'll lower the deficit and ask the ultra-wealthy and corporations to pay their fair share. And no one making under $400,000 per year will pay a penny more in taxes."
    prediction_results = model.run(sample_text)

    print("Generated text: ", prediction_results)

    # Saving model
    model.save(args.output_dir)

    if args.tgis:

        # Load model in TGIS
        # HACK: export args.output_dir as MODEL_NAME for TGIS
        # container to pick up automatically
        os.environ["MODEL_DIR"] = os.path.dirname(args.output_dir)
        os.environ["MODEL_NAME"] = os.path.join(
            "models", os.path.basename(args.output_dir), "artifacts"
        )

        loaded_model = load_model(is_distributed=True, model_path=args.output_dir)

    else:
        # Use trained model directly
        loaded_model = model

    ## Evaluation
    print_colored("[Starting Evaluation]")

    validation_stream = dataset_info.dataset_loader()[1]

    print_colored("Getting model predictions...")
    truncate_input_tokens = args.max_source_length + args.max_target_length
    predictions, references = get_model_preds_and_references(
        loaded_model, validation_stream, truncate_input_tokens, args.max_target_length
    )

    export_model_preds(args.preds_file, predictions, validation_stream)

    metric_funcs = [SUPPORTED_METRICS[metric_name] for metric_name in args.metrics]
    print_colored("Metrics to be calculated: {}".format(args.metrics))

    for metric_func in metric_funcs:
        metric_res = metric_func(predictions=predictions, references=references)
        print_colored(metric_res)
